#include <QWidget>
#include <QVBoxLayout>
#include <QLineEdit>
#include <QPushButton>
#include <QTextEdit>
#include <QLabel>
#include <QString>
#include <QThread>
#include <QCheckBox>
#include <QHBoxLayout>
#include <cmath>
#include <QSpinBox>
#include <QApplication>

using namespace std;
using namespace std::chrono;

class SortWorker : public QObject {
    Q_OBJECT
public:
    SortWorker(vector<int>& arr, unsigned int n, bool individualTask) 
        : arr(arr), n(n), individualTask(individualTask), shouldStop(false) {}
    
    void stop() {
        shouldStop = true;
    }
    
public slots:
    void performSort() {
        auto start_time = high_resolution_clock::now();
        unsigned int step = n / 100;
        if (!step) step = 1;
        unsigned int next_report = step;
        bool first_output = true;
        // Сортування за спаданням для індивідуального завдання, за зростанням - для звичайного
        std::function<bool(int, int)> comp = [](int a, int b) { return a < b; };
        
        for (unsigned int i = 0; i < n - 1; i++) {
            if (shouldStop) {
                emit sortingCancelled();
                return;
            }
            for (unsigned int gap = arr.size() / 2; gap > 0; gap /= 2) {
                for (unsigned int i = gap; i < arr.size(); i++) {
                    int temp = arr[i];
                    unsigned int j;
                    for (j = i; j >= gap && arr[j - gap] > temp; j -= gap) {
                        arr[j] = arr[j - gap];
                    }
                    arr[j] = temp;
                }
            }
            
            auto current_time = high_resolution_clock::now();
            auto elapsed_time = duration_cast<seconds>(current_time - start_time);
            
            if (i + 1 == next_report || (elapsed_time.count() >4 && first_output)) {
                unsigned short progress = ((i + 1) * 100) / n;
                
                unsigned int current = i + 1;
                if (current > n) current = n;
                
                emit progressUpdated(progress, current, n);
                
                next_report += step;
                first_output = false;
            }
        }
        
        auto stop = high_resolution_clock::now();
        auto duration = duration_cast<microseconds>(stop - start_time);
        
        emit progressUpdated(100, n, n);
        emit sortingFinished(duration.count(), is_sorted(arr.data(), n, comp));
    }
    
signals:
    void progressUpdated(int progress, unsigned int current, unsigned int total);
    void sortingFinished(int duration, bool sorted);
    void sortingCancelled();
    
private:
    vector<int>& arr;
    unsigned int n;
    bool individualTask;
    std::atomic<bool> shouldStop;
    template<typename Compare>
    bool is_sorted(int* arr, unsigned int n, Compare comp) {
        for (unsigned int i = 0; i < n - 1; i++) {
            if (comp(arr[i+1], arr[i])) return false;
        }
        return true;
    }
};
class SortWorkerWorse : public QObject {
    Q_OBJECT
public:
    SortWorkerWorse(vector<vector<float>>& arr, unsigned int n, bool individualTask) 
        : arr(arr), n(n), individualTask(individualTask), shouldStop(false) {}
    
    void stop() {
        shouldStop = true;
    }
    
public slots:
    void performSort() {
        auto start_time = high_resolution_clock::now();
        unsigned int step = n / 100;
        if (!step) step = 1;
        unsigned int next_report = step;
        bool first_output = true;
        // Сортування за спаданням для індивідуального завдання, за зростанням - для звичайного
        std::function<bool(float, float)> comp = [](float a, float b) { return a < b; };
        
        for (unsigned int i = 0; i < n - 1; i++) {
            if (shouldStop) {
                emit sortingCancelled();
                return;
            }
            for (int gap = arr[0].size() / 2; gap > 0; gap /= 2) {
                for (int i = gap; i < arr[0].size(); i++) {
                    float tempMax = arr[0][i];
                    unsigned int colIndex = i;
                    for (int j = i; j >= gap; j -= gap) {
                        if (arr[0][j - gap] > tempMax) {
                            tempMax = arr[0][j - gap];
                            colIndex = j - gap;
                        }
                    }
                    if (colIndex != i) for (unsigned int j = 0; j < arr.size(); ++j)swap(arr[j][i], arr[j][colIndex]);
                }
            }
            
            auto current_time = high_resolution_clock::now();
            auto elapsed_time = duration_cast<seconds>(current_time - start_time);
            
            if (i + 1 == next_report || (elapsed_time.count() >4 && first_output)) {
                unsigned short progress = ((i + 1) * 100) / n;
                
                unsigned int current = i + 1;
                if (current > n) current = n;
                
                emit progressUpdated(progress, current, n);
                
                next_report += step;
                first_output = false;
            }
        }
        
        auto stop = high_resolution_clock::now();
        auto duration = duration_cast<microseconds>(stop - start_time);
        
        emit progressUpdated(100, n, n);
        emit sortingFinished(duration.count(), is_sorted(arr, n, comp));
    }
    
signals:
    void progressUpdated(int progress, unsigned int current, unsigned int total);
    void sortingFinished(int duration, bool sorted);
    void sortingCancelled();
    
private:
    vector<vector<float>>& arr;
    unsigned int n;
    bool individualTask;
    std::atomic<bool> shouldStop;
    template<typename Compare>
    bool is_sorted(std::vector<vector<float>>, unsigned int n, Compare comp) {
        for (unsigned int i = 0; i < n - 1; i++) {
            if (comp(arr[0][i+1], arr[0][i])) return false;
        }
        return true;
    }
};

class SortingApp : public QWidget {
    Q_OBJECT
public:
    SortingApp(QWidget *parent = nullptr) : QWidget(parent), worker(nullptr), wworker(nullptr), isSorting(false) {
        setupUI();
    }

    ~SortingApp() {
        if (workerThread.isRunning()) {
            workerThread.quit();
            workerThread.wait();
        }
    }

private:
    QSpinBox *inputField;
    QPushButton *sortButton;
    QTextEdit *outputField;
    QCheckBox *showArrayCheckBox;
    vector<int> arr;
    vector<vector<float>> arrI;
    QThread workerThread;
    SortWorker *worker;
    SortWorkerWorse *wworker;
    QCheckBox *individualTaskCheckBox;
    bool isSorting;

    void setupUI() {
        QVBoxLayout *mainLayout = new QVBoxLayout(this);

        QLabel *label = new QLabel("Скільки чисел сортувати?", this);
        mainLayout->addWidget(label);

        inputField = new QSpinBox(this);
        inputField->setRange(1, 10000000);
        inputField->setValue(10);  // Встановлюємо початкове значення
        mainLayout->addWidget(inputField);

        sortButton = new QPushButton("Сортувати", this);
        connect(sortButton, &QPushButton::clicked, this, &SortingApp::onSortButtonClicked);
        mainLayout->addWidget(sortButton);

        QHBoxLayout *checkBoxLayout = new QHBoxLayout();
        
        showArrayCheckBox = new QCheckBox("Показати масив до та після сортування", this);
        checkBoxLayout->addWidget(showArrayCheckBox);

        individualTaskCheckBox = new QCheckBox("Індивідуальне завдання", this);
        checkBoxLayout->addWidget(individualTaskCheckBox);

        mainLayout->addLayout(checkBoxLayout);

        outputField = new QTextEdit(this);
        outputField->setReadOnly(true);
        mainLayout->addWidget(outputField);

        setLayout(mainLayout);
        setWindowTitle("Сортування масиву");
        resize(500, 300);
    }

    void sortArray() {
        arr.clear();
        arrI.clear();
        if (individualTaskCheckBox->isChecked()){
            arrI.resize(inputField->value());
            for (unsigned int i = 0; i < inputField->value(); i++){
                arrI[i].resize(inputField->value());
                for (unsigned int j = 0; j < inputField->value(); j++) arrI[i][j] = rand()/10000.0-100000;
            }
        }
        else{
        arr.resize(inputField->value());
        for (unsigned int i = 0; i < inputField->value(); i++) arr[i] = rand() % 20000001 - 10000000;
        }
        outputField->clear();
        if (showArrayCheckBox->isChecked()) {
            outputField->append("Масив до сортування:");
            if (individualTaskCheckBox->isChecked()) outputField->append(vvToString(arrI));
            else outputField->append(arrayToString(arr));
        }
        if (individualTaskCheckBox->isChecked()){
            wworker = new SortWorkerWorse(arrI, inputField->value(), individualTaskCheckBox->isChecked());
            wworker->moveToThread(&workerThread);
            outputField->append("==Режим індивідуального завдання #7==\n");
            for (int b=0;b<inputField->value();b++){
            int min=0;
            for (int a=1;a<inputField->value();a++)if (min>arrI[b][a])min=a;
            arrI[b][min]=log(fabs(arrI[b][min]));}
            connect(this, &SortingApp::startSort, wworker, &SortWorkerWorse::performSort);
            connect(wworker, &SortWorkerWorse::progressUpdated, this, &SortingApp::updateProgress);
            connect(wworker, &SortWorkerWorse::sortingFinished, this, &SortingApp::finishSorting);
            connect(&workerThread, &QThread::finished, wworker, &QObject::deleteLater);
        }else{
            worker = new SortWorker(arr, inputField->value(), individualTaskCheckBox->isChecked());
            worker->moveToThread(&workerThread);
            connect(this, &SortingApp::startSort, worker, &SortWorker::performSort);
            connect(worker, &SortWorker::progressUpdated, this, &SortingApp::updateProgress);
            connect(worker, &SortWorker::sortingFinished, this, &SortingApp::finishSorting);
            connect(&workerThread, &QThread::finished, worker, &QObject::deleteLater);}
        workerThread.start();
        emit startSort();
    }
    
    QString arrayToString(const vector<int>& arr) {
        QString result;
        for (size_t i = 0; i < arr.size(); ++i) {
            if (individualTaskCheckBox->isChecked())result += QString::number(arr[i], 'f', 2); else result += QString::number(static_cast<int>(arr[i]));
            if (i < arr.size() - 1) result += ", ";
        }
        return result;
    }
    QString vvToString(const vector<vector<float>>& arr) {
        QString result;
        for (size_t i = 0; i < arr.size(); ++i){
            for (size_t j = 0; j < arr.size(); ++j) {
            if (individualTaskCheckBox->isChecked())result += QString::number(arr[i][j], 'f', 2); else result += QString::number(static_cast<float>(arr[i][j]));
            if (j < arr.size() - 1) result += ", ";
        }
            if (i < arr.size() - 1) result+='\n';
        }
        return result;
    }

private slots:
    void onSortButtonClicked() {
        if (!isSorting) {
        isSorting = true;
        sortButton->setText("Скасувати");
            sortArray();
        } else resetSortingState();
    }

    void updateProgress(int progress, unsigned int current, unsigned int total) {
        outputField->append(QString("Відсортовано %1% (%2 з %3 елементів)").arg(progress).arg(current).arg(total));
    }

    void finishSorting(int duration, bool sorted) {
        outputField->append(QString("\nЧас сортування: %1 мікросекунд").arg(duration));
        outputField->append(sorted ? "Масив відсортовано правильно." : "Помилка: масив не відсортовано правильно.");
        
        if (showArrayCheckBox->isChecked()) {
            outputField->append("\nМасив після сортування:");
            if (individualTaskCheckBox->isChecked())outputField->append(vvToString(arrI));else
            outputField->append(arrayToString(arr));
            arr.clear();
        }
        
        resetSortingState();
    }
    void resetSortingState() {
        isSorting = false;
        sortButton->setText("Сортувати");
        if(worker){
            worker->disconnect();worker->deleteLater();}
        if(wworker){wworker->deleteLater();
            wworker->disconnect();}
        worker=nullptr;
        wworker=nullptr;
    }

signals:
    void startSort();
};

int main(int argc, char *argv[]) {
    QApplication app(argc, argv);
    SortingApp sortingApp;
    sortingApp.show();
    return app.exec();
}
#include "main.moc"
